// joystick.h -- Class declaration for joystick object -*-C++-*-
#ifndef INC_JOYSTICK
#define INC_JOYSTICK

// Standard X11 header files
#include <X11/Intrinsic.h>
#include <X11/Xlib.h>

// X Input Extention header files
#include <X11/extensions/XInput.h>


// Class definition
class Joystick {
public:
  Joystick (void);		// Initialize the joystick class object
  ~Joystick ();			// Release the joystick object
  int initialize (Display *);	// Open the joystick device
				// Returns 1 on success, 0 on failure

  int present (void) {		// Returns True if the joystick is active
    return (device != NULL);
  }

  int x_direction (void);	// Get the current relative x direction
				//  (-1 for left, 0 for center, 1 for right)
  double x_position (void);	// Get the current absolute x position
				//  (-1.0 <= x <= 1.0, or 0.0 if centered)
  int y_direction (void);	// Get the current relative y direction
				//  (-1 for up, 0 for center, 1 for down)
  double y_position (void);	// Get the current absolute y position
				//  (-1.0 <= x <= 1.0, or 0.0 if centered)
  int buttons (void);		// Get the current button state
				//  (bit 0 = 1st button)

private:
  void get_stick (XValuatorState **); // Get the raw joystick data
  void calibrate (void);	// Calibrate the joystick

private:
  Display *display;		// Display the joystick is connected to
  XDevice *device;		// Joystick device
  int x_min, x_max;		// Extents of x values
  int xc_min, xc_max;		// Extents of x when centered
  int y_min, y_max;		// Extents of y values
  int yc_min, yc_max;		// Extents of y when centered
};

#endif // INC_JOYSTICK
