// An Adventure Game
// from A.N.A.L.O.G. Magazine number 10
// Atari version by Michael Duboy
// X port by Trevin Beattie


// Public header files
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>


// Private header files
#include "data.h"
#include "transient.h"
#include "xgui.h"


// Forward declarations
void init_game (void);
void play_game (void);
void swim (void);
void treasure_square (void);
void take_loan (void);
void buy_food (void);
void hire_companion (void);
void fight_bandits (int how_many, const char *who, Boolean treasure);
void show_status (void);
void show_score (void);


// Program main()
// This routine initializes the game and keeps it moving in a loop
// until the game is over and the player doesn't want to play again.
// It then exits.  Simple?
int
main (int argc, char **argv)
{
  init_x ();
  do {
    init_game ();
    show_window ();
    redraw_map ();
    update_status ();
    do {
      play_game ();
    } while (!game_over);
    show_score ();
  } while (!game_over);
  close_x ();
  return 0;
}


// init_game()
// Initialize data fields in the GameState structure for a new game.
void
init_game (void)
{
  int x, i, j;
  char b[25];

  srandom ((unsigned int) time (NULL));
  game_data.LoanedMoney = False;
  game_data.Interest = 0;
  game_data.PartySize = 1;
  game_data.DebtToCompanion = 0;
  game_data.Gold = 3;
  game_data.Difficult = False;
  game_data.Food = 6;
  game_data.Water = 3;
  game_data.HitPoints = 100 + random () % 100;
  game_data.HitPointsCompanion = 0;
  game_data.X = 1;
  game_data.Y = 0;

  // Generate a random sequence of codes for the treasure chests.
  for (i = 0; i < 25; i++)
    game_data.TreasureCodes[i] = i + 1;
  for (i = 10; i < 20; i += 2)
    for (j = 10; j < 20; j += 2)
      map_data[i][j].type = TREASURE;
  for (x = 0; x < 10; x++)
    {
      // Pick a spot in the middle of the string (excluding end 5 bytes).
      i = 5 + random () % 15;
      // 'Cut the deck' at this point
      memcpy (&b[0], &game_data.TreasureCodes[0], i);
      memcpy (&game_data.TreasureCodes[0],
	      &game_data.TreasureCodes[i], 25 - i);
      memcpy (&game_data.TreasureCodes[25 - i], &b[0], i);
      // Double-cut it in the following order: 12345 becomes 34125.
      memcpy (&b[0], &game_data.TreasureCodes[10], 10);
      memcpy (&b[10], &game_data.TreasureCodes[0], 10);
      memcpy (&b[20], &game_data.TreasureCodes[20], 5);
      memcpy (&game_data.TreasureCodes[0], &b[0], 25);
    }

  // Force an update of the map.
  update_map ();
}


// play_game()
// Move the player across the field,
// and take action depending on where he goes.
void
play_game (void)
{
  int dx, dy;
  int bandits;
  int r;

  // Wait until the player moves.
  move_event_loop (&dx, &dy);
  if (game_over)
    return;

  // Make sure we stay within bounds.
  if ((game_data.X + dx < 0) || (game_data.X + dx > 19))
    dx = 0;
  if ((game_data.Y + dy < 0) || (game_data.Y + dy > 19))
    dy = 0;

  // Check whether we've hit a wall.
  if (map_data[game_data.X + dx][game_data.Y + dy].type == WALL)
    {
      dx = 0;
      dy = 0;
      // Ouch!
      --game_data.HitPoints;
      show_status ();
      if (game_over)
	return;
    }

  // Make the move
  if (dx || dy)
    {
      game_data.Food -= 2 * game_data.PartySize;
      game_data.Water -= game_data.PartySize;
      show_status ();
      if (game_over)
	return;
      game_data.X += dx;
      game_data.Y += dy;
      update_map ();
    }

  // Check for increased difficulty.
  if (!game_data.Difficult && (game_data.Gold > 5000))
    {
      game_data.Difficult = True;
      update_map ();
    }

  // Branch off depending on where the player went to.
  switch (map_data[game_data.X][game_data.Y].type)
    {
    case ROAD:
      // On the road, you have a 1%-20% chance of meeting bandits,
      // depending on how far you are from the city.
      r = random () % 100 + 1 + game_data.Y;
      if (r > 99)
	{
	  bandits = r / 10;
	  fight_bandits (r, "bandits", FALSE);
	}
      break;
    case FOREST:
      // In the forest, you have a 20%-40% chance of meeting bandits,
      // depending on how far you are from the city.
      r = random () % 100 + 1 + game_data.Y;
      if (r > 80)
	{
	  bandits = r / 10 + game_data.Y;
	  fight_bandits (r, "bandits", FALSE);
	}
      break;
    case SEA:
      swim ();
      break;
    case LOOSE_ROAD:
      r = random () % 100 + 1 + game_data.Y;
      if (game_data.Difficult)
	{
	  // This area has gone underwater
	  swim ();
	} else {
	  if (r > 99)
	    {
	      bandits = r / 10;
	      fight_bandits (r, "bandits", FALSE);
	    }
	}
      break;
    case LOOSE_FOREST:
      r = random () % 100 + 1 + game_data.Y;
      if (game_data.Difficult)
	{
	  // This area has gone underwater
	  swim ();
	} else {
	  if (r > 80)
	    {
	      bandits = r / 10 + game_data.Y;
	      fight_bandits (r, "bandits", FALSE);
	    }
	}
      break;
    case TREASURE:
      treasure_square ();
      break;
    case BANK:
      if (!game_data.LoanedMoney)
	take_loan ();
      break;
    case SUPPLIES:
      buy_food ();
      break;
    case COMPANION:
      if (game_data.PartySize == 1)
	hire_companion ();
      break;
    case FINISH:
      if (game_data.Gold >= 200)
	game_over = True;
      break;
    }
}


// swim()
// Player is moving through water.  Try to drown him.
void
swim (void)
{
  int r;

  r = random () % 100 + 1 + game_data.Y;
  if (game_data.PartySize == 2)
    {
      // At sea, your companion has a 60%-72% chance of drowning at each step.
      if (r > 40)
	{
	  transient->print (TM_CLEAR, "Companion has drowned.");
	  transient->popup ();
	  game_data.PartySize = 1;
	  show_status ();
	  if (game_over)
	    return;
	  timed_event_loop (2000);
	  transient->hide ();
	}
    } else {
      // At sea, you have a 40%-52% chance of drowning at each step.
      if (r > 60)
	{
	  transient->print (TM_CLEAR, "You've drowned!");
	  game_over = True;
	  timed_event_loop (3000);
	  transient->hide ();
	}
    }
}


// treasure_square()
// Give the player a chance to win some money
void
treasure_square (void)
{
  int r, x, a, nguesses = 0;
  const char *who;

  a = game_data.TreasureCodes[map_data[game_data.X][game_data.Y].data];

  transient->print (TM_CLEAR, "Treasure square!");
  transient->popup ();
  while (1)
    {
      transient->print (TM_ADD, "Do you wish to go to another square? [Y/N]");
      if (yes_no_event_loop () || game_over)
	{
	  transient->hide ();
	  return;
	}

      r = random () % 100 + 1;
      if (r > 84 - nguesses)
	{
	  transient->print (TM_CLEAR, "Evil spirits!");
	  game_data.Water = game_data.Water * 9 / 10;
	  game_data.Food = game_data.Food * 9 / 10;
	  show_status ();
	  if (game_over)
	    return;
	}

      x = number_event_loop ("What is your guess? %d", 1, 25);
      if (game_over)
	return;
      nguesses++;
      if (x == a)
	break;
      if (game_data.PartySize == 1)
	continue;
      if (x > a)
	transient->print (TM_ADD, "Companion says to go lower.");
      else
	transient->print (TM_ADD, "Companion says to go higher.");
    }

  transient->print (TM_CLEAR, "The stone has opened!");
  switch (random () % 3)
    {
    case 0: who = "Goblins"; break;
    case 1: who = "Orcs"; break;
    case 2: who = "Ghouls"; break;
    }
  r = random () % 100 + 1;
  // display_bandits ();
  fight_bandits (r, who, TRUE);
  map_data[game_data.X][game_data.Y].type = FOREST;
}


// take_loan()
// Offer to loan some money to the player.
void
take_loan (void)
{
  int r;
  char sbuf[32];

  transient->print (TM_CLEAR, "Greetings!  I'm Abdul.");
  transient->print (TM_ADD, "Need a loan? [Y/N]");
  transient->popup ();
  if (yes_no_event_loop ())
    {
      r = random () % 100 + 51;
      sprintf (sbuf, "I offer you %d gold peices.", r);
      transient->print (TM_REPLACE, sbuf);
      sprintf (sbuf, "Repayment will be %d.", r * 2);
      transient->print (TM_ADD, sbuf);
      transient->print (TM_ADD, "Do you accept my terms? [Y/N]");
      if (yes_no_event_loop ())
	{
	  game_data.Gold += r;
	  game_data.Interest = r * 2;
	  game_data.LoanedMoney = True;
	  show_status ();
	}
    }
  transient->hide ();
}


// buy_food ()
// Sell food and water to the player.
void
buy_food (void)
{
  int n;

  transient->print (TM_CLEAR, "Welcome to Kwik-E Mart!  I am Apu.");
  transient->print (TM_ADD, "How much food would you like?");
  transient->popup ();
  n = number_event_loop ("%d days worth", 0, (game_data.Gold - 1) / 2);
  if (game_over)
    return;
  game_data.Food += n * 6;
  game_data.Gold -= n * 2;
  show_status ();
  transient->print (TM_CLEAR, "How much water would you like?");
  n = number_event_loop ("%d days worth", 0, 999);
  game_data.Water += n * 3;
  if (n)
    game_data.Gold -= 1;
  show_status ();
  transient->hide ();
}


// hire_companion()
// Give the player somebody else to help fight and take blows.
void
hire_companion (void)
{
  int r;
  char sbuf[64];

  transient->print (TM_CLEAR, "Do you need a companion? [Y/N]");
  transient->popup ();
  if (yes_no_event_loop ())
    {
      r = (random () % 100 + 1) * (250 - game_data.HitPoints);
      sprintf (sbuf, "Your companion wants %d in gold to be payed", r);
      transient->print (TM_CLEAR, sbuf);
      transient->print (TM_ADD, "to him or his heir after the adventure.");
      transient->print (TM_ADD, "Do you accept? [Y/N]");
      if (yes_no_event_loop ())
	{
	  game_data.PartySize = 2;
	  game_data.DebtToCompanion = r;
	  game_data.HitPointsCompanion = (r / 50 + game_data.HitPoints) / 2;
	  show_status ();
	}
    }
  transient->hide ();
}


// fight_bandits()
//
void
fight_bandits (int how_many, const char *who, Boolean treasure)
{
  int killed = 0;
  int fighter = game_data.PartySize;
  int r, r2;
  char sbuf[80];

  transient->print (TM_CLEAR, "Want to run? [Y/N]");
  transient->popup ();
  if (!yes_no_event_loop ())
    {
      while (!game_over)
	{
	  if (how_many <= 0)
	    {
	      r2 = random () % 100 + 1;
	      if (treasure)
		r = (game_data.X + 1) * r2 * r2 / (game_data.Y + 1);
	      else
		r = 0;
	      r += r2 * game_data.X * game_data.Y / 10;
	      sprintf (sbuf, "You've found %d in gold on the %s", r, who);
	      transient->print (TM_CLEAR, sbuf);
	      game_data.Gold += r;
	      show_status ();
	      if (!game_over)
		timed_event_loop (3000);
	      break;
	    }
	  if (fighter == 1)
	    {
	      game_data.HitPoints -= random () % how_many + 1;
	      show_status ();
	      if (game_over)
		break;
	      killed = random () % game_data.HitPoints;
	      if (killed > how_many)
		killed = how_many;
	      how_many -= killed;
	      sprintf (sbuf, "There are %d %s; you've killed %d %s",
		       how_many, who, killed, who);
	      transient->print (TM_CLEAR, sbuf);
	      timed_event_loop (1500);
	      fighter = game_data.PartySize;
	    } else {
	      game_data.HitPointsCompanion -= random () % how_many + 1;
	      show_status ();
	      if (game_over)
		break;
	      if (game_data.PartySize == 2)
		{
		  killed = random () % game_data.HitPointsCompanion;
		  if (killed > how_many)
		    killed = how_many;
		  how_many -= killed;
		  sprintf (sbuf,
			   "There are %d %s; your companion has killed %d %s",
			   how_many, who, killed, who);
		  transient->print (TM_CLEAR, sbuf);
		  timed_event_loop (1500);
		}
	      fighter = 1;
	    }
	}
    }
  transient->hide ();
}


// display_status()
// Check the player's status information; end the game if dead,
// otherwise show the new status.
void
show_status (void)
{
  int weight;
  char msgbuf[128];
  int n;
  Boolean t_was_mapped = transient->mapped;

  update_status ();

  // Find out how much weight the party is carrying.
  // The original formula is: 1 food = 1/6, 1 gold = 1/1000, 1 water = 1/2
  // Each member of the party can carry 100 units.
  // The weights are scaled *3000 for fast integer arithmetic.
  while ((weight = (game_data.Food * 500 + game_data.Gold * 3
		    + game_data.Water * 1500))
	 > 300000 * game_data.PartySize)
    {
      sprintf (msgbuf, "%d.%d is too much weight!",
	       weight / 3000, ((weight + 150) / 300) % 10);
      transient->print (TM_CLEAR, msgbuf);
      transient->popup ();

      n = number_event_loop ("Throw away how much gold? %d",
			     0, game_data.Gold);
      game_data.Gold -= n;
      update_status ();

      n = number_event_loop ("Throw away how much food? %d",
			     0, game_data.Food);
      game_data.Food -= n;
      update_status ();

      n = number_event_loop ("Throw away how much water? %d",
			     0, game_data.Water);
      game_data.Water -= n;
      update_status ();
      if (t_was_mapped)
	transient->hide ();
    }

  // Check whether the player has died.
  if (game_data.Water < 0)
    {
      transient->print (TM_CLEAR, "Died from lack of water");
      transient->popup ();
      timed_event_loop (3000);
      game_over = True;
      if (t_was_mapped)
	transient->hide ();
      return;
    }
  else if (game_data.Food < 0)
    {
      transient->print (TM_CLEAR, "Starved to death");
      transient->popup ();
      timed_event_loop (3000);
      game_over = True;
      if (t_was_mapped)
	transient->hide ();
      return;
    }
  else if (game_data.HitPoints < 1)
    {
      transient->print (TM_CLEAR, "Died from wounds");
      transient->popup ();
      timed_event_loop (3000);
      game_over = True;
      if (t_was_mapped)
	transient->hide ();
      return;
    }

  // Check whether the companion has died.
  if ((game_data.HitPointsCompanion < 1) && (game_data.PartySize == 2))
    {
      transient->print (TM_CLEAR, "Companion is dead.\n");
      transient->popup ();
      game_data.HitPointsCompanion = 0;
      update_status ();
      timed_event_loop (2000);
      if (t_was_mapped)
	transient->hide ();
      game_data.PartySize = 1;
      game_data.Gold -= game_data.DebtToCompanion;
      game_data.DebtToCompanion = 0;
    }

  update_status ();
}


// show_score()
// Display the final game status
// and ask the player if he wants to play again.
void
show_score (void)
{
  game_data.Gold -= game_data.DebtToCompanion + game_data.Interest;
  game_data.DebtToCompanion = game_data.Interest = 0;
  game_data.PartySize = 1;
  update_status ();
  transient->print (TM_CLEAR, "Your rating is:");
  if (game_data.Gold > 20000)
    transient->print (TM_ADD, "OVERGORD");
  else if (game_data.Gold > 10000)
    transient->print (TM_ADD, "GORDENT");
  else if (game_data.Gold > 2000)
    transient->print (TM_ADD, "AVENTOR");
  else if (game_data.Gold >= 0)
    transient->print (TM_ADD, "HINDMOST");
  else
    transient->print (TM_CLEAR, "You've been put in debtors' prison!");
  transient->print (TM_ADD, "Do you want to play again? [Y/N]");
  transient->popup ();
  game_over = False;
  game_over = !yes_no_event_loop ();
  transient->hide ();
}
