// Class object for the map window

// Public include files
#include <iostream.h>
#include <stdlib.h>

// Private include files
#include "data.h"
#include "map.h"

// Color xpm pixmaps
#include "bank.xpm"
#include "chest.xpm"
#include "companion.xpm"
#include "finish.xpm"
#include "hero.xpm"
#include "road.xpm"
#include "sea.xpm"
#include "store.xpm"
#include "tree.xpm"
#include "wall-e.xpm"
#include "wall-n.xpm"
#include "wall-s.xpm"
#include "wall-w.xpm"


// Data structures
MapWindow *map_window;


// Class methods for the map window

// Initialize the map window
MapWindow::MapWindow (Display *d, Window parent)
{
  XGCValues gcv;
  XSetWindowAttributes wattr;

  display = d;

  // Set the values we will need for the graphics context.
  if (!XAllocNamedColor (display, XDefaultColormapOfScreen
			 (XDefaultScreenOfDisplay (display)),
			 "brown", &bgcolor, &bgcolor))
    {
      cerr << "xANALOGadventure: failed to allocate brown.\n";
      exit (-1);
    }
  gcv.background = bgcolor.pixel;
  gcv.subwindow_mode = IncludeInferiors;

  // Create the map window.
  map_window = XCreateSimpleWindow (display, parent, 0, 0,
				    MAP_WIDTH, MAP_HEIGHT, 0,
				    gcv.background, gcv.background);
  if (!map_window)
    {
      cerr << "xANALOGadventure: failed to create a window.\n";
      exit (-1);
    }

  // Create the hero's subwindow.
  hero_subwindow = XCreateSimpleWindow (display, map_window,
					game_data.X * ITEM_WIDTH,
					game_data.Y * ITEM_HEIGHT,
					ITEM_WIDTH, ITEM_HEIGHT, 0,
					gcv.background, gcv.background);
  if (!hero_subwindow)
    {
      cerr << "xANALOGadventure: failed to create a window.\n";
      exit (-1);
    }
  // Select button events on the map window.
  XSelectInput (display, map_window,
		ButtonPressMask | ButtonReleaseMask | ExposureMask);
  XSelectInput (display, hero_subwindow, ExposureMask);

  // Create our graphics context.
  gc = XCreateGC (display, map_window, GCBackground | GCSubwindowMode, &gcv);
  if (!gc)
    {
      cerr << "xANALOGadventure: failed to create a GC.\n";
      exit (-1);
    }

  // Create pixmaps to use for map squares.
  xpmattr.valuemask = XpmExactColors | XpmRGBCloseness;
  xpmattr.exactColors = False;
  xpmattr.red_closeness = 20000;
  xpmattr.green_closeness = 10000;
  xpmattr.blue_closeness = 40000;
  if (XpmCreatePixmapFromData (display, map_window, hero_xpm,
			       &icon[IHERO].pix, &icon[IHERO].mask,
			       &xpmattr) < 0)
    {
      cerr << "xANALOGadventure: failed to create the hero's icon.\n";
      exit (-1);
    }
  if (XpmCreatePixmapFromData (display, map_window, bank_xpm,
			       &icon[IBANK].pix, &icon[IBANK].mask,
			       &xpmattr) < 0)
    {
      cerr << "xANALOGadventure: failed to create the bank icon.\n";
      exit (-1);
    }
  if (XpmCreatePixmapFromData (display, map_window, chest_xpm,
			       &icon[ITREASURE].pix, &icon[ITREASURE].mask,
			       &xpmattr) < 0)
    {
      cerr << "xANALOGadventure: failed to create the treasure icon.\n";
      exit (-1);
    }
  if (XpmCreatePixmapFromData (display, map_window, companion_xpm,
			       &icon[ICOMPANION].pix,
			       &icon[ICOMPANION].mask, &xpmattr) < 0)
    {
      cerr << "xANALOGadventure: failed to create the companion icon.\n";
      exit (-1);
    }
  if (XpmCreatePixmapFromData (display, map_window, finish_xpm,
			       &icon[IFINISH].pix, &icon[IFINISH].mask,
			       &xpmattr) < 0)
    {
      cerr << "xANALOGadventure: failed to create the finish icon.\n";
      exit (-1);
    }
  if (XpmCreatePixmapFromData (display, map_window, road_xpm,
			       &icon[IROAD].pix, &icon[IROAD].mask,
			       &xpmattr) < 0)
    {
      cerr << "xANALOGadventure: failed to create the road icon.\n";
      exit (-1);
    }
  if (XpmCreatePixmapFromData (display, map_window, sea_xpm,
			       &icon[ISEA].pix, &icon[ISEA].mask,
			       &xpmattr) < 0)
    {
      cerr << "xANALOGadventure: failed to create the sea icon.\n";
      exit (-1);
    }
  if (XpmCreatePixmapFromData (display, map_window, store_xpm,
			       &icon[ISUPPLIES].pix, &icon[ISUPPLIES].mask,
			       &xpmattr) < 0)
    {
      cerr << "xANALOGadventure: failed to create the store icon.\n";
      exit (-1);
    }
  if (XpmCreatePixmapFromData (display, map_window, tree_xpm,
			       &icon[IFOREST].pix, &icon[IFOREST].mask,
			       &xpmattr) < 0)
    {
      cerr << "xANALOGadventure: failed to create the forest icon.\n";
      exit (-1);
    }
  if (XpmCreatePixmapFromData (display, map_window, wall_e_xpm,
			       &icon[IWALL_E].pix, &icon[IWALL_E].mask,
			       &xpmattr) < 0)
    {
      cerr << "xANALOGadventure: failed to create the east wall icon.\n";
      exit (-1);
    }
  if (XpmCreatePixmapFromData (display, map_window, wall_n_xpm,
			       &icon[IWALL_N].pix, &icon[IWALL_N].mask,
			       &xpmattr) < 0)
    {
      cerr << "xANALOGadventure: failed to create the north wall icon.\n";
      exit (-1);
    }
  if (XpmCreatePixmapFromData (display, map_window, wall_s_xpm,
			       &icon[IWALL_S].pix, &icon[IWALL_S].mask,
			       &xpmattr) < 0)
    {
      cerr << "xANALOGadventure: failed to create the south wall icon.\n";
      exit (-1);
    }
  if (XpmCreatePixmapFromData (display, map_window, wall_w_xpm,
			       &icon[IWALL_W].pix, &icon[IWALL_W].mask,
			       &xpmattr) < 0)
    {
      cerr << "xANALOGadventure: failed to create the west wall icon.\n";
      exit (-1);
    }
}


// Release the map window
MapWindow::~MapWindow ()
{
  int i;

  // Free the pixmaps created by xpm
  for (i = 0; i < N_ICONS; i++)
    {
      XFreePixmap (display, icon[i].pix);
      XFreePixmap (display, icon[i].mask);
    }
  // Free the colors allocated by xpm
  XFreeColors (display, XDefaultColormapOfScreen
	       (XDefaultScreenOfDisplay (display)),
	       xpmattr.pixels, xpmattr.npixels, 0);
  // Release the xpm attributes
  XpmFreeAttributes (&xpmattr);
  // Release the graphics context
  XFreeGC (display, gc);
  // Release the windows
  XDestroyWindow (display, hero_subwindow);
  XDestroyWindow (display, map_window);
}


// Map the map window
void
MapWindow::show (void)
{
  XMapWindow (display, hero_subwindow);
  XMapWindow (display, map_window);
}


// Redraw the entire map window
void
MapWindow::redraw (void)
{
  redraw_area (map_window, 0, 0, MAP_WIDTH, MAP_HEIGHT);
}

// Redraw a selected part of the map window
void
MapWindow::redraw_area (Window target_window,
			int clip_x, int clip_y, int clip_w, int clip_h)
{
  int x, y, pic;
  XColor fg_color, dummy_color;
  XPoint pts[8];

  // If the target window is the hero,
  // adjust the clipping area to match the map window.
  if (target_window == hero_subwindow)
    {
      clip_x += game_data.X * ITEM_WIDTH;
      clip_y += game_data.Y * ITEM_HEIGHT;
    }
  // If the target window is not one of ours, ignore the redraw request.
  else if (target_window != map_window)
    return;

  // If we will be redrawing the square our hero is on,
  // be sure to clear the area first.
  if ((game_data.X * ITEM_WIDTH < clip_x + clip_w)
      && ((game_data.X + 1) * ITEM_WIDTH > clip_x)
      && (game_data.Y * ITEM_HEIGHT < clip_y + clip_h)
      && ((game_data.Y + 1) * ITEM_HEIGHT > clip_y))
    {
      XSetForeground (display, gc, bgcolor.pixel);
      XSetClipMask (display, gc, None);
      XFillRectangle (display, map_window, gc, game_data.X * ITEM_WIDTH,
		      game_data.Y * ITEM_HEIGHT, ITEM_WIDTH, ITEM_HEIGHT);
    }

  // Redraw the map area
  for (y = 0; y < 20; y++) {
    for (x = 0; x < 20; x++) {
      // Figure out whether we can skip this block
      if (((x + 1) * ITEM_WIDTH <= clip_x)
	  || (x * ITEM_WIDTH >= clip_x + clip_w)
	  || ((y + 1) * ITEM_HEIGHT <= clip_y)
	  || (y * ITEM_HEIGHT >= clip_y + clip_h))
	continue;
      switch (map_data[x][y].type)
	{
	case LOOSE_ROAD:
	  pic = game_data.Difficult ? ISEA : IROAD;
	  break;
	case ROAD:
	  pic = IROAD;
	  break;
	  XAllocNamedColor (display, XDefaultColormapOfScreen
			    (XDefaultScreenOfDisplay (display)),
			    "brown", &fg_color, &dummy_color);
	  XSetForeground (display, gc, fg_color.pixel);
	  XFillRectangle (display, map_window, gc, x * ITEM_WIDTH,
			  y * ITEM_HEIGHT, ITEM_WIDTH, ITEM_HEIGHT);
	  break;
	case LOOSE_FOREST:
	  pic = game_data.Difficult ? ISEA : IFOREST;
	  break;
	case FOREST:
	  pic = IFOREST;
	  break;
	  XAllocNamedColor (display, XDefaultColormapOfScreen
			    (XDefaultScreenOfDisplay (display)),
			    "brown", &fg_color, &dummy_color);
	  XSetForeground (display, gc, fg_color.pixel);
	  XFillRectangle (display, map_window, gc, x * ITEM_WIDTH,
			  y * ITEM_HEIGHT, ITEM_WIDTH, ITEM_HEIGHT);
	  XAllocNamedColor (display, XDefaultColormapOfScreen
			    (XDefaultScreenOfDisplay (display)),
			    "green", &fg_color, &dummy_color);
	  XSetForeground (display, gc, fg_color.pixel);
	  // Create a `tree'
	  pts[7].x = pts[0].x = x * ITEM_WIDTH + (ITEM_WIDTH / 2);
	  pts[7].y = pts[0].y = y * ITEM_HEIGHT;
	  pts[1].x = (x + 1) * ITEM_WIDTH - 1;
	  pts[6].x = x * ITEM_WIDTH;
	  pts[6].y = pts[5].y = pts[2].y = pts[1].y
	    = y * ITEM_HEIGHT + ((ITEM_HEIGHT * 2) / 3);
	  pts[3].x = pts[2].x = x * ITEM_WIDTH + ((ITEM_WIDTH * 3) / 5);
	  pts[5].x = pts[4].x = x * ITEM_WIDTH + ((ITEM_WIDTH * 2) / 5);
	  pts[4].y = pts[3].y = (y + 1) * ITEM_HEIGHT - 1;
	  XFillPolygon (display, map_window, gc, &pts[0], 8,
			Nonconvex, CoordModeOrigin);
	  break;
	case SEA:
	  pic = ISEA;
	  break;
	  XAllocNamedColor (display, XDefaultColormapOfScreen
			    (XDefaultScreenOfDisplay (display)),
			    "blue", &fg_color, &dummy_color);
	  XSetForeground (display, gc, fg_color.pixel);
	  XFillRectangle (display, map_window, gc, x * ITEM_WIDTH,
			  y * ITEM_HEIGHT, ITEM_WIDTH, ITEM_HEIGHT);
	  break;
	case WALL:
	  switch (map_data[x][y].data)
	    {
	    case 'E': pic = IWALL_E; break;
	    case 'N': pic = IWALL_N; break;
	    case 'S': pic = IWALL_S; break;
	    case 'W': pic = IWALL_W; break;
	    default: pic = ISEA; break;
	    }
	  break;
	  XAllocNamedColor (display, XDefaultColormapOfScreen
			    (XDefaultScreenOfDisplay (display)),
			    "green", &fg_color, &dummy_color);
	  XSetForeground (display, gc, fg_color.pixel);
	  XFillRectangle (display, map_window, gc, x * ITEM_WIDTH,
			  y * ITEM_HEIGHT, ITEM_WIDTH, ITEM_HEIGHT);
	  XAllocNamedColor (display, XDefaultColormapOfScreen
			    (XDefaultScreenOfDisplay (display)),
			    "gray50", &fg_color, &dummy_color);
	  XSetForeground (display, gc, fg_color.pixel);
	  break;
	case TREASURE:
	  pic = ITREASURE;
	  break;
	  XAllocNamedColor (display, XDefaultColormapOfScreen
			    (XDefaultScreenOfDisplay (display)),
			    "yellow", &fg_color, &dummy_color);
	  XSetForeground (display, gc, fg_color.pixel);
	  XFillRectangle (display, map_window, gc, x * ITEM_WIDTH,
			  y * ITEM_HEIGHT, ITEM_WIDTH, ITEM_HEIGHT);
	  break;
	case BANK:
	  pic = IBANK;
	  break;
	case SUPPLIES:
	  pic = ISUPPLIES;
	  break;
	case COMPANION:
	  pic = ICOMPANION;
	  break;
	case FINISH:
	  pic = IFINISH;
	  break;
	}
      XSetClipMask (display, gc, icon[pic].mask);
      XSetClipOrigin (display, gc, x * ITEM_WIDTH, y * ITEM_HEIGHT);
      XCopyArea (display, icon[pic].pix, map_window, gc, 0, 0,
		 ITEM_WIDTH, ITEM_HEIGHT, x * ITEM_WIDTH, y * ITEM_HEIGHT);
    }
  }

  // Redraw the hero
  XSetClipMask (display, gc, icon[IHERO].mask);
  XSetClipOrigin (display, gc, 0, 0);
  XCopyArea (display, icon[IHERO].pix, hero_subwindow, gc, 0, 0,
	     ITEM_WIDTH, ITEM_HEIGHT, 0, 0);
}


// Force a redraw when the player moves
void
MapWindow::update (void)
{
  XExposeEvent event;

  XUnmapWindow (display, hero_subwindow);
  // This causes the area previously under the hero to be redrawn.
  event.type = Expose;
  event.serial = 0;
  event.send_event = True;
  event.display = display;
  event.window = map_window;
  event.x = game_data.X * ITEM_WIDTH;
  event.y = game_data.Y * ITEM_HEIGHT;
  event.width = ITEM_WIDTH;
  event.height = ITEM_HEIGHT;
  event.count = 0;
  XSendEvent (display, map_window, True, ExposureMask, (XEvent *) &event);
  XMoveWindow (display, hero_subwindow,
	       game_data.X * ITEM_WIDTH, game_data.Y * ITEM_HEIGHT);
  XMapWindow (display, hero_subwindow);
}


// Force a redraw of the entire map
void
MapWindow::clean (void)
{
  XExposeEvent event;
  XUnmapWindow (display, map_window);
  XMapWindow (display, map_window);
}


// Figure out the relative direction of a button click in the map area
void
MapWindow::get_offset (XButtonEvent *event, int *dx, int *dy)
{
  int bx, by;

  // If the mouse click did not occur within the map window, ignore it.
  if (event->window != map_window)
    return;

  // If the mouse was clicked on a direct horizontal or vertical
  // path from the hero, move in that direction.
  if (event->x < game_data.X * ITEM_WIDTH)
    bx = event->x - game_data.X * ITEM_WIDTH;
  else if (event->x >= (game_data.X + 1) * ITEM_WIDTH)
    bx = event->x - (game_data.X + 1) * ITEM_WIDTH;
  else
    {
      *dx = 0;
      if (event->y < game_data.Y * ITEM_HEIGHT)
	*dy = -1;
      else if (event->y >= (game_data.Y + 1) * ITEM_HEIGHT)
	*dy = 1;
      else
	*dy = 0;
      return;
    }

  if (event->y < game_data.Y * ITEM_HEIGHT)
    by = event->y - game_data.Y * ITEM_HEIGHT;
  else if (event->y >= (game_data.Y + 1) * ITEM_HEIGHT)
    by = event->y - (game_data.Y + 1) * ITEM_HEIGHT;
  else
    {
      *dy = 0;
      *dx = (bx < 0) ? -1 : 1;
      return;
    }

  // If the mouse was clicked within 1 square of the hero,
  // we are going directly to that square.
  if ((bx >= -ITEM_WIDTH) && (bx < ITEM_WIDTH)
      && (by >= -ITEM_HEIGHT) && (by < ITEM_HEIGHT))
    {
      *dx = (bx < 0) ? -1 : 1;
      *dy = (by < 0) ? -1 : 1;
      return;
    }

  // For spots further out, the direction depends on the octant in which
  // the button was pressed, measured as approximately 22.5 degrees from
  // the nearest axis of the nearest corner of the hero's square.
  if (bx == 0)
    {
      // Special case, right octant
      *dx = ((by == -1) || (by == 0)) ? 1 : 0;
      *dy = (by < 0) ? -1 : 1;
      return;
    }
  // 100000*tan(22.5) =  41421.xxx
  // 100000*tan(67.5) = 241421.xxx
  by = 100000 * by / bx;
  if (bx < 0)
    {
      // A left octant
      if (by >= 41421)
	{
	  *dy = -1;
	  *dx = (by > 241421) ? 0 : -1;
	}
      else if (by <= -41421)
	{
	  *dy = 1;
	  *dx = (by < -241421) ? 0 : -1;
	}
      else
	{
	  *dy = 0;
	  *dx = -1;
	}
    }
  else if (bx > 0)
    {
      // A right octant
      if (by <= -41421)
	{
	  *dy = -1;
	  *dx = (by < -241421) ? 0 : 1;
	}
      else if (by >= 41421)
	{
	  *dy = 1;
	  *dx = (by > 241421) ? 0 : 1;
	}
      else
	{
	  *dy = 0;
	  *dx = 1;
	}
    }
  return;

  // Figure out the `square' the button was clicked in.
  bx = event->x / ITEM_WIDTH;
  by = event->y / ITEM_HEIGHT;
  if (bx < game_data.X)
    *dx = -1;
  else if (bx > game_data.X)
    *dx = 1;
  if (by < game_data.Y)
    *dy = -1;
  else if (by > game_data.Y)
    *dy = 1;
}


// update_map()
// Redraw part of the map when the player moves
void
update_map (void)
{
  map_window->update ();
}

// redraw_map()
// Redraw all of the map
void
redraw_map (void)
{
  map_window->clean ();
}
