// Class object for transient message window


// Public include files
#include <iostream.h>
#include <string.h>
#include <stdlib.h>

// Private include files
#include "transient.h"
#include "xgui.h"


// Data structures
TransientMessages *transient;


// Class methods for the transient window

// Initialize the transient window
TransientMessages::TransientMessages (Display *d, Window parent) {
  XGCValues gcv;
  int maxwidth, maxheight;

  num_messages = 0;

  display = d;

  // Figure out what font we're going to use.
  font = XLoadQueryFont (display,
			 "-adobe-utopia-medium-r-*-*-18-*-*-*-p-*-iso8859-1");
  if (!font) {
    // Our preferred font doesn't exist; revert to a fixed font.
    font = XLoadQueryFont (display, "fixed");
    if (!font) {
      cerr << "xANALOGadventure: failed to allocate a font.\n";
      exit (-1);
    }
  }

  // Set the values we will need for the graphics context.
  gcv.foreground = BlackPixel (display, DefaultScreen (display));
  gcv.background = WhitePixel (display, DefaultScreen (display));
  gcv.font = font->fid;

  // Make sure our window is big enough for any text we might print.
  winheight = 8 + 3 * (font->ascent + font->descent);
  winwidth = 8 +		// I think this is the longest string we have:
    XTextWidth (font,
		"There are 288 bandits; your companion has killed 88 bandits.",
		60);

  // Create our window.
  window = XCreateSimpleWindow (display, parent, (MAP_WIDTH - winwidth) / 2,
				(MAP_HEIGHT - winheight) / 2, winwidth,
				winheight, 2, gcv.foreground, gcv.background);

  // Select the event types we will use
  XSelectInput (display, window, ExposureMask | KeyPressMask);

  // Create our graphics context.
  gc = XCreateGC (display, window, GCForeground | GCBackground | GCFont, &gcv);
}


TransientMessages::~TransientMessages () {
  int i;

  // Release the graphics context
  XFreeGC (display, gc);
  // Release the window
  XDestroyWindow (display, window);

  for (i = 0; i < num_messages; i++) {
    // Release the message strings
    delete message[i];
  }
}


// Display the message window
void TransientMessages::popup (void) {
  XMapRaised (display, window);
  mapped = True;
}


// Hide the message window
void TransientMessages::hide (void) {
  XUnmapWindow (display, window);
  mapped = False;
}


// Add text to the message window
void TransientMessages::print (int how, const char *line) {
  int i;

  // Make room for the new line.
  switch (how) {
  case TM_CLEAR:
    for (i = 0; i < num_messages; i++)
      delete message[i];
    num_messages = 0;
    break;
  case TM_ADD:
    if (num_messages == 3) {
      delete message[0];
      message[0] = message[1];
      msglen[0] = msglen[1];
      linewidth[0] = linewidth[1];
      message[1] = message[2];
      msglen[1] = msglen[2];
      linewidth[1] = linewidth[2];
      --num_messages;
    }
    break;
  case TM_REPLACE:
    if (num_messages) {
      --num_messages;
      delete message[num_messages];
    }
    break;
  }

  // Allocate storage for the line and copy it.
  msglen[num_messages] = strlen (line);
  message[num_messages] = new char [msglen[num_messages] + 1];
  strcpy (message[num_messages], line);
  linewidth[num_messages] = XTextWidth (font, message[num_messages],
					msglen[num_messages]);
  num_messages++;

  // If the window is already popped up, redraw it.
  if (mapped)
    XClearArea (display, window, 0, 0, 0, 0, True);
}


// Draw the transient window
void TransientMessages::redraw (void) {
  int lineheight, i;

  lineheight = font->ascent + font->descent;
  for (i = 0; i < num_messages; i++) {
    XDrawImageString (display, window, gc, (winwidth - linewidth[i]) / 2,
		      (winheight - num_messages * lineheight) / 2
		      + i * lineheight + font->ascent, message[i], msglen[i]);
  }
}
