// Functions and data used for the X interface  -*-C++-*-

#ifndef INC_XGUI
#define INC_XGUI


// Public include files
#include <signal.h>
#include <X11/Intrinsic.h>


// Private include files
#include "joystick.h"
#include "map.h"


// define's
#define NUM_STATUS_LINES 3

// Class definition for the X interface
class XGUI {
public:
  XGUI (void);
  ~XGUI ();

  void show (void);		// Map the windows
  void hide (void);		// Unmap the windows
  void redraw (void);		// Redraw the entire window
  void redraw_area (Window W, int x, int y,
		    int w, int h); // Redraw part of the window
  XEvent *next_event (void);	// Grab the next non-management event
  void event_pause (int timeout); // Process events for a period of time
  void signal (int);		// Send a signal event to ourself
  void noise (void);		// Make a noise
  void print (const char *);	// Print a line of text in the status area
  void clear_status (void);	// Clear the status area

private:
  Display *display;
  Window window;
  GC gc;
  XFontStruct *font;
  Joystick joystick;
  struct sigaction old_handler; // For SIGALRM handler
  unsigned int status_height;
  int num_status_lines;
  char *status_line[NUM_STATUS_LINES];
  int status_len[NUM_STATUS_LINES], line_width[NUM_STATUS_LINES];
};


// Forward declarations
void init_x (void);		// Initialize the X Window structures
void show_window (void);	// Map the windows
void close_x (void);		// Release the X Window structures
void move_event_loop (int *dx, int *dy);
				// Process X events until movement is detected
Boolean yes_no_event_loop (void);
				// Process X events until 'yes'
				//  or 'no' is answered
int number_event_loop (const char *message, int min, int max);
				// Process X events until a number is selected
void timed_event_loop (int ms);	// Process X events for a period of time
void update_status (void);	// Redraw the status window


#endif
